/* -*- mode: c; c-file-style: "gnu" -*- */

#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>

#include "schoenhage.h"

/* Introduction
 *
 * Consider this (non-working) gcd algorithm, with input two numbers a
 * and b, both of size 2n bits. Split them into halves,
 *
 *  a = a_1 2^n + a_2, b = b_1 2^n + b_2
 *
 * Compute, recursively, r1 = gcd(a1, b1), and two numbers u1 and v1
 * such that
 *
 *   r1 = u1 a1 + v1 b1                                        (*)
 *
 * Form the corresponding linear combination of the original numbers,
 *
 *   r  = u1  a + v1  b
 *
 * No matter where u1 and v1 came from, r is always a multiple of
 * gcd(a,b). We get
 *
 *   r  = u1 (a1 2^ n + a2) + v1 (b1 2^n + b2)
 *      = r1 2^n + u1 a2 + v1 b2
 * 
 * The nice thing about this expression is that the first term is
 * considerably smaller than our original numbers. So we could try
 * finding the gcd (or a smaller multiple therof) by recursively
 * computing
 *
 *   gcd(r, b mod r)
 *
 * This gives us a sub-quadratic algorithm for gcd computation.
 * However, this doesn't quite work, and the reason is that r isn't
 * small. The terms u1 a2 and v1 b2 are about the same size of our
 * original numbers, so we don't get anywhere.
 *
 * To get a working algorithm, we need to choose r1, u1 and v1
 * differently. We still want equation (*) to hold, but with numbers
 * r1, u1 and v1 of about the same size, n/2 bits.
 *
 * Such numbers exist: Consider the extended euclidian algorithm
 * applied to a1 and b1, and take a remaider r1 halfway through that
 * computation, and the corresponding u1 and u2.
 *
 * We then get an
 *
 *   r  = r1 2^n + u1 a2 + v1 b2
 *
 * of size 3n/2, 3/4 of the size of the numbers we started with. And
 * that's enough to get an efficient algorithm.
 *
 * So the heart of Schnhage's algorithm is the function hgcd ("half
 * gcd"), which takes two numbers a and b as input, of size 2n bits
 * each, and recursively computes four numbers u, v, u' and v',
 * corresponding to two numbers
 *
 *   r  = u a + v b
 *   r' = u'a + v'b
 *
 * of size close to n, which corresponds to remainders close to the middle of
 * the euclidean remainder sequence for a and b.
 */

/* Notation: The input for hgcd is r_0 = a, r_1 = b, where a > b > 0.
 *
 * Define a remainder and quotient sequence
 *
 *   r_{i-2} = q_{i-1} r_{i-1} + r_i, 2 <= i <= k
 *    ...
 *   r_{k-1} = q_k r_k
 *
 * To each i, 1 <= i <= k, associate a matrix 
 *
 *          / 0   1  \
 *   Q_i = |          |
 *          \ 1 -q_i /
 *
 * Define the products
 *
 *   R_i = Q_i Q_{i-1} ... Q_1
 *
 * We're interested in the R_j, with j chosen so that
 *
 *   #r_i > #r_0 / 2 iff i <= j
 *
 * where #x denotes the size of x in bits. hgcd computes this matrix.
 *
 * We will denote the elements of R with R = (u, v; u', v').
 */

/* The trick: Let
 *
 *   a = a_1 2^k + a_2
 *   b = b_1 2^k + b_2
 *
 * with a_2, b_2 < 2^k. Also assume that a_1 <= (b_1)^2 (or perhaps
 * instead the stronger assumption that #a_1 < 2 #b_1, <==> #b_1 >=
 * #a_1 / 2 - 1?). Form the full quotients a/b, and truncated
 * quotients a_1/b_1:
 *
 *     a =   q   b +   r
 *   a_1 = q_1 b_1 + r_1
 *
 * Substitution the decompositions of a and b into the first equation,
 * and then substituting the second, yields
 *
 *   2^k b_1 (q -q_1) = r_1 2^k + a_2 - q b_2 - r
 *      < r_1 2^k + 2^k - q b_2 - r
 *     <= b_1 2^k - q b_2 - r
 *
 * Divide, to get
 *
 *     q - q_1 < 1 - (q b_2 + r) / (2^k b_1)
 * ==> q - q_1 <= 0
 *
 * Let q_1 = q + d. Then
 *
 *   - 2^k b_1 d = r_1 2^k + a_2 - q b_2 - r
 *
 *   2^k b_1 d = q b_2 + r - r_1 2^k - a_2
 *     <= q b_2 + r = (q_1 - d) b_2 + r
 *
 * Now use a_1 <= b_1^2, which implies q_1 <= a_1 / b_1 <= b_1 and
 *
 *   b d = 2^k b_1 d + b_2 d <= q_1 b_2 + r
 *       < b_1 2^k + b <= 2b
 *
 * and we conclude that d < 2, i.e. that d is zero or one. One can
 * show (Aho) that d != 0 can happen only for the final hgcd step. So
 * we can check for this, and compensate. (XXX How?)
 *
 * When truncating, we also have trouble with errors in the
 * computation of the remainders. Assuming that we have q = q_1, we
 * get an error, according to Aho, of at most
 *
 *   |r - 2^k r_1| <= 2^k(1+q)
 *
 * It's not clear if that matters; the reduction (a;b) -> R (a;b) can
 * never lose any factors of the gcd, but errors in R could in
 * principle introduce spurious factors. What matters for the
 * algorithm is that the reduction halves the size of the input, and
 * it seems reasonable that an R computed from truncated input will
 * suffice to clear many of the upper bits.
 */

void
trace(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  gmp_vfprintf(stderr, format, args);
  va_end(args);
}

int
normalize(mpz_t r, mpz_t a, mpz_t b)
{
  int sa, sb;

  /* Ensure a >= b > 0 */
  sa = mpz_sgn(a);
  sb = mpz_sgn(b);

  if (sa < 0)
    mpz_neg(a, a);
  if (sb < 0)
    mpz_neg(b, b);

  if (!sa)
    {
      if (!sb)
	abort();

      mpz_set(r, b);
      return 0;
    }
  else if (!sb)
    {
      mpz_set(r, a);
      return 0;
    }

  if (mpz_cmp(a, b) < 0)
    mpz_swap(a, b);

  return 1;
}

void
schoenhage_array_init(struct schoenhage_array *R)
{
  mpz_init(R->m[0][0]);
  mpz_init(R->m[0][1]);
  mpz_init(R->m[1][0]);
  mpz_init(R->m[1][1]);
}

void
schoenhage_array_clear(struct schoenhage_array *R)
{
  mpz_clear(R->m[0][0]);
  mpz_clear(R->m[0][1]);
  mpz_clear(R->m[1][0]);
  mpz_clear(R->m[1][1]);
}

void
schoenhage_array_identity(struct schoenhage_array *R)
{
  /* Initialize with the identity */
  mpz_set_ui(R->m[0][0], 1);
  mpz_set_ui(R->m[0][1], 0);
  mpz_set_ui(R->m[1][0], 0);
  mpz_set_ui(R->m[1][1], 1);
}

void
schoenhage_array_copy(struct schoenhage_array *R,
		      const struct schoenhage_array *S)
{
  unsigned i, j;
  for (i = 0; i<2; i++)
    for (j = 0; j<2; j++)
      mpz_set(R->m[i][j], S->m[i][j]);
}

int
schoenhage_array_equal(const struct schoenhage_array *R,
		       const struct schoenhage_array *S)
{
  unsigned i, j;

  for (i = 0; i<2; i++)
    for (j = 0; j<2; j++)
      if (mpz_cmp(R->m[i][j], S->m[i][j]))
	return 0;
  
  return 1;
}

/* Returns tru if the matrices have one row in common */
int
schoenhage_array_row_equal(const struct schoenhage_array *R,
			   const struct schoenhage_array *S)
{
  if (schoenhage_array_equal(R, S))
    return 1;
  
  if (!mpz_cmp(R->m[0][0], S->m[1][0])
      && !mpz_cmp(R->m[0][1], S->m[1][1]))
    {
      trace("schoenhage_array_row_equal: +\n");
      return 1;
    }
  else if (!mpz_cmp(R->m[1][0], S->m[0][0])
	   && !mpz_cmp(R->m[1][1], S->m[0][1]))
    {
      trace("schoenhage_array_row_equal: -\n");
      return 1;
    }

  return 0;
}

void
schoenhage_array_step_ui(struct schoenhage_array *R,
			 mp_limb_t q)
{
  /* Update R from the left
   *
   *   0  1   A B   =   C      D
   *   1 -q   C D      A-qC   B-qD
   */
  unsigned i;
  mpz_t t;
  mpz_init(t);
  for (i = 0; i<2; i++)
    {
      mpz_mul_ui(t, R->m[1][i], q);
      mpz_sub(R->m[0][i], R->m[0][i], t);
      mpz_swap(R->m[0][i], R->m[1][i]);
    }
  mpz_clear(t);
}

void
schoenhage_array_step_left(struct schoenhage_array *R,
			   const mpz_t q)
{
  /* Update R from the left
   *
   *   0  1   A B   =   C      D
   *   1 -q   C D      A-qC   B-qD
   */
  unsigned i;
  mpz_t t;
  mpz_init(t);
  for (i = 0; i<2; i++)
    {
      mpz_mul(t, R->m[1][i], q);
      mpz_sub(R->m[0][i], R->m[0][i], t);
      mpz_swap(R->m[0][i], R->m[1][i]);
    }
  mpz_clear(t);
}

/* (y1; y2) = R (x1 ; x2) */
void
schoenhage_array_vector(mpz_t y1, mpz_t y2,
			const struct schoenhage_array *R,
			const mpz_t x1, const mpz_t x2)
{
  /* Allows xk and yk to overlap */
  mpz_t t;
  mpz_t a;

  mpz_init(t);
  mpz_init(a);

  mpz_mul(a, R->m[0][0], x1);
  mpz_mul(t, R->m[0][1], x2);
  mpz_add(t, t, a);
  
  mpz_mul(a, R->m[1][0], x1);
  mpz_mul(y2, R->m[1][1], x2);
  mpz_add(y2, y2, a);

  mpz_set(y1, t);

  mpz_clear(a);
  mpz_clear(t);
}

/* R = A B */
void
schoenhage_array_mult(struct schoenhage_array *R,
		      const struct schoenhage_array *A,
		      const struct schoenhage_array *B)
{
  unsigned i;
  assert(R != A);
  assert(B != A);

  for (i = 0; i<2; i++)
    schoenhage_array_vector(R->m[0][i], R->m[1][i],
			    A,
			    B->m[0][i], B->m[1][i]);
}

/* Product of matrices (0, 1; 1, -q) */
void
egcd(struct schoenhage_array *R, unsigned m, mp_limb_t a, mp_limb_t b)
{
  /* #x <= m  iff  x < 2^m */
  mp_limb_t limit = ((mp_limb_t)(1)) << m;

#if 0
  trace("egcd: size = %d, a = %lu, b = %lu\n",
	size, (long) a, (long) b);
#endif
  assert(a > b);
  assert(a >= limit);
  
  schoenhage_array_identity(R);

  /* We need to return an R corresponding to a large r and small r1.
     Through out this loop, (a;b) = R (original_a;original_b) */
  while (b >= limit)
    { /* The same as |b| >= size */
      mp_limb_t q = a / b;
      mp_limb_t r = a % b;

      schoenhage_array_step_ui(R, q);
      
      /* trace("(q,r) = (%d, %d) ", q, r); */
            
      /* Euclidean step */
      a = b; b = r;
    }
  assert(a >= limit);
  assert(b < limit);
  
  /* trace("\n"); */
}


void
fix(struct schoenhage_array *R, mpz_t x, mpz_t y)
{
  /* We have (x;y) = R (a; b). We modify y and R, if needed,
   *
   * to ensure that x > y > 0. */

  /* Jabelean's criterion tells us that x is correct.
   *
   * Let k = 2n, A = a 2^k + A', B = b 2^k + B'
   *
   * Define quotient sequences:
   *
   * A_0 = A, A_1 = B, Q_{i-1} = floor(A_i/A_{i-1})
   * a_0 = a, a_1 = b, q_{i-1} = floor(a_i/a_{i-1})
   *
   * Then, if
   *
   *   a_{i+2} >= 2^n for i = 1, 2, ..., k,
   *
   * then
   *
   *   Q_i = q_i,              for i <= k,
   *   (U_i, V_i) = (u_i,v_i), for i <= k+1
   *
   * and we can compute
   *
   *   A_k     = u_k A + v_k B
   *   A_{k+1} = u_{k+1} A + v_{k+1} B
   */
   
  /* However, this does not help us. We can aim for hgcd to return
   * pairs a_k. a_{k+1}, of size larger than m1, and any fix number of
   * steps before a_i drops to a size smaller than m1. But then we
   * construct A_k and A_{k+1}, we may find that we get to A_i smaller
   * than m sooner than required, and we can't reconstruct the earlier
   * A_{k-1}.
   *
   * Instead, we only insist that r and r' are of the desired size,
   * and allow some spurious factors to get into the gcd. These are
   * aliminated at the end, for example using lehmer's algorithm.
   */
  
  assert(mpz_sgn(x) > 0);

  if (mpz_sgn(y) < 0)
    {
      /* Or second */
      trace("fix: y < 0\n");
      mpz_neg(R->m[1][0], R->m[1][0]);
      mpz_neg(R->m[1][1], R->m[1][1]);
      mpz_neg(y,y);
    }

  if (mpz_cmp(x, y) < 0)
    {
      /* Swap rows */
      trace("fix: x < y\n");
      mpz_swap(R->m[0][0], R->m[1][0]);
      mpz_swap(R->m[0][1], R->m[1][1]);
      mpz_swap(x,y);
    }
}

/* More precisely, the specification for this function is as follows:
   Compute a matrix (u,v;u',v'), corresponding to two values r = u a +
   v b, r'= u'a + v'b, in the remainder sequence of a,b. Let r'' = r
   mod r'. Then we should have have #r > m, #r' > m, and #r'' <= m,
   with m = ceil(#a / 2). For this to be possible at all, we must have
   #b > m in the input.

   Furthermore, we prefer to have gcd(u,v) = gcd(u',v') = 1, or at
   least these pairs should have only small common factors. */

void
check_hgcd(const struct schoenhage_array *R,
	   const mpz_t a, const mpz_t b)
{
  /* Check that the returned values match the spec. */
  unsigned n;
  unsigned m;
  
  mpz_t t;
  mpz_t r;
  mpz_t r1;
  
  mpz_init(t);
  mpz_init(r);
  mpz_init(r1);

  n = mpz_sizeinbase(a, 2);
  m = (n + 1) / 2;
    
  schoenhage_array_vector(r, r1, R, a, b);
  if (mpz_sizeinbase(r, 2) <= m)
    {
      trace("check_hgcd: r too small (%d bits, expected > %d)\n",
	    mpz_sizeinbase(r, 2), m);
    fail:
      trace_array(" R", R);
      trace("a = %Zd, b = %Zd,\n"
	    "  r = %Zd, r1 = %Zd\n", a, b, r, r1);
      
      abort();
    }

  if (mpz_sizeinbase(r1, 2) > m)
    {
      trace("check_hgcd: r1 too large (%d bits, expected <= %d)\n"
	    "  r1 = %Zd\n",
	    mpz_sizeinbase(r1, 2), m, r1);
      goto fail;
    }
  
  mpz_gcd(t, R->m[0][0], R->m[0][1]);
  if (mpz_cmp_ui(t, 1) != 0)
    trace("check_hgcd: Nontrivial gcd(u,v): %Zd\n", t);

  mpz_gcd(t, R->m[1][0], R->m[1][1]);
  if (mpz_cmp_ui(t, 1) != 0)
    trace("check_hgcd: Nontrivial gcd(u',v'): %Zd\n", t);

  /* If there were no truncation errors, we would have
   *
   *   u' <= b/r, v' <= a/r
   *
   * Now we just get the estimate #u', #v' <= n - m = floor(n/2).
   
   * This bound also needs a proof. */

  if (mpz_sizeinbase(R->m[1][0], 2) > n - m)
    {
      trace("check_hgcd: u' unexpectedly large\n");
      trace("  got %d bits, expected <= %d\n",
	    mpz_sizeinbase(R->m[1][0], 2), n - m);
    }

  if (mpz_sizeinbase(R->m[1][1], 2) > n - m)
    {
      trace("check_hgcd: v' unexpectedly large\n");
      trace("  got %d bits, expected <= %d\n",
	    mpz_sizeinbase(R->m[1][1], 2), n - m);
    }
  
  mpz_clear(t);
  mpz_clear(r);
  mpz_clear(r1);
}

/* Analysis of the sizes of the numbers in the hgcd algorithm.
 *
 * Start with a > b > 0,
 *
 *   n = #a, k = floor(n/2) (this parameter is not the same as m =
 *   ceil(n/2). Let
 *
 *   a1 = floor(a/2^k), b1 = floor(b/2^k)
 *   a1 >= b1 >= 0
 *
 * The first recursive call gives us a matrix R1 with
 *
 *   r1  = u1 a1 + v1 b1, #r1  > ceil(#a1 / 2)
 *   r1' = u1'a1 + v1'b1, #r1' > ceil(#a1 / 2)
 *   r1'' = r1 mod r1', #r1'' <= ceil(#a1 / 2)
 *
 * We also have
 *
 *   abs(u1') <= a1/r1
 *   #u1' <= #a1 - ceil(#a1 / 2) = floor(#a1 / 2)
 *
 * and similarly for v1', u1 and v1.
 *
 * Next we compute (x;y) = R (a;b), we have
 *
 *   x = u1 (a1 2^k + a2) + v1 (b1 2^k + b2)
 *     = r1 2^k + u1 a2 + v1 b2
 *
 *   y = u1' (a1 2^k + a2) + v1'(b1 2^k + b2)
 *     = r1' 2^k + u1' a2 + v1' b2   (+ x somtimes)
 *
 *   z = x mod y
 *
 * We have
 *
 *   #(r1 2^k) > ceil(#a1 / 2) + k
 *   #(u a2) <= floor(#a1 / 2) + k < #(r1 2^k)
 *
 * Since u1' and v1' have different signs, also
 *
 *   #(u1' a2 + v1' b2) < #(r1 2^k)
 *
 * Therefore,
 *
 *   #x >= k + ceil(#a1 / 2)     = k + ceil( (n-k)/2 )
 *   #y >= k + ceil(#a1 / 2)     = k + ceil( (n-k)/2 )
 *
 * FIXME: THis bound isn't quite right.
 *
 *   #z <= k + ceil(#a1 / 2) + 1 = k + ceil( (n-k)/2 ) + 1
 *
 * After one reduction, z = x mod y, we get y > z > 0 This costs us
 * one factor q in R. With k = floor(n/2), the bounds are
 *
 *   #x >= floor(n/2) + ceil(n/4) > m
 *   #y >= floor(n/2) + ceil(n/4) > m
 *   #z <= floor(n/2) + ceil(n/4) + 1 = m + ceil(n/4) + 1
 *
 * Next, choose another parameter l, depending on the actual size of
 * y. We'll see how in a moment. Set
 *
 *   y1 = floor(y/2^l), z1 = floor(z/2^k)
 *   y1 >= z1 >= 0
 *
 * Recurse again, giving us a matrix R2, with
 *
 *   r2  = u2 y1 + v2 z1, #r2   > ceil(#y1 / 2)
 *   r2' = u2'y1 + v2'z1, #r2'  > ceil(#y1 / 2)
 *   r2'' = r2 mor r2',   #r2'' <= ceil(#y1 / 2)
 *
 * Compute, in the same way as x and y,
 *
 *   r = u2 (y1 2^l + y2) + v2 (z1 2^l + z2)
 *     = r2 2^l + u2 y2 + v2 z2
 *
 *   r' = u2' (y1 2^l + y2) + v2' (z1 2^l + z2)
 *      = r2' 2^l + u2 y2 + v2' z2
 *
 * We get, in the same way,
 *
 *   #r   >= l + ceil(#y1 / 2)    = l + ceil( (#y - l) / 2)
 *   #r'  >= l + ceil(#y1 / 2)    = l + ceil( (#y - l) / 2)
 *   #r'' <= l + ceil(#y1 / 2)    = l + ceil( (#y - l) / 2) + 1
 *
 * So what is l? l should be chosen so that l + ceil ( (#y - l) / 2) >
 * m, as there is no way to recover from too small r, r'. The right
 * choice is
 *
 *   l = 2m - #y + 1
 *
 * Then #y - l = 2 (#y - m) - 1
 *
 *   #r' <= 2m - #y + 1 + ceil( (2 (#y - m) - 1)/2) + 1
 *        = m + 2
 *
 * So #r' may still be too large, but at most by two bits. If it is
 * too large, we do a few more euclidean steps, corresponding to a
 * matrix Q2, to get it to the right size.
 *
 * The final output is then R = Q2 R2 Q R1.
 */
 
/* Must be called with a >= b > 0, a > 0. No modification of a and b.
 * Returns a matrix R such that (r;r') = R * (a;b) gives two numbers
 * of size about 3/4 that of a. */

void
hgcd(struct schoenhage_array *R, const mpz_t a, const mpz_t b)
{
  unsigned n = mpz_sizeinbase(a, 2);
  unsigned m = (n + 1) / 2;
  
  static int entry = 0;
  int count = ++entry;
  
  assert(mpz_sgn(a) > 0);

  if (mpz_sizeinbase(b, 2) <= m)
    {
      trace("hgcd(%d): b small, b = %Zd\n", count, b);
      schoenhage_array_identity(R);
      return;
    }
  if (mpz_cmp(a, b) == 0)
    {
      trace("hgcd(%d): a = b = %Zd\n", count, a);
      
      schoenhage_array_identity(R);
      schoenhage_array_step_ui(R, 1);
      return;
    }

  assert(mpz_cmp(a, b) > 0);

  /* We really need a better base case, probably based on lehmer. */
  if (mpz_size(a) == 1)
    {
      assert(mpz_size(b) <= 1);
      egcd(R, m, mpz_getlimbn(a, 0), mpz_getlimbn(b, 0));
#if 0
      trace("hgcd(%d): egcd, R = [ %Zd, %Zd\n"
	    "                      %Zd, %Zd ]\n", count,
	    R->m[0][0], R->m[0][1], R->m[1][0], R->m[1][1]);
#endif
      return;
    }

  {
    struct schoenhage_array S;

    /* Copies that we reduce as we go */
    mpz_t x;
    mpz_t y;
    	  
    mpz_t x1;
    mpz_t y1;
    mpz_t q;
    mpz_t r;

    unsigned k;
    unsigned l;
    unsigned n1;
    
    schoenhage_array_init(&S);
    mpz_init_set(x, a);
    mpz_init_set(y, b);
    
    mpz_init(x1);
    mpz_init(y1);
    mpz_init(q);
    mpz_init(r);

    /* First split off floor(asize/2) bits */
    k = n / 2;

    assert(mpz_sizeinbase(x, 2) > m);
    
    mpz_fdiv_q_2exp(x1, x, k);
    mpz_fdiv_q_2exp(y1, y, k);

    assert(mpz_sizeinbase(x1, 2) == n - k);

    hgcd(R, x1, y1);
    check_hgcd(R, x1, y1);
    
    schoenhage_array_vector(x, y, R, x, y);
#if 0
    trace("hgcd(%d) reduced: x = %Zd, y = %Zd\n", count, x, y);

    trace("hgcd(%d): expected: %d, got: %d, %d\n",
	  count, a1size + a1size/2,
	  mpz_sizeinbase(x,2), mpz_sizeinbase(y,2));
#endif
    fix(R, x, y);
    
    assert(mpz_sizeinbase(x, 2) > m);
    if (mpz_sizeinbase(y, 2) <= m)
      goto done;
    
    mpz_fdiv_qr(q, r, x, y);
    mpz_swap(r, x);
    mpz_swap(x, y);
    
    schoenhage_array_step_left(R, q);

    assert(mpz_sizeinbase(x, 2) > m);
    if (mpz_sizeinbase(y, 2) <= m)
      goto done;
    
    /* New smaller size */
    n1 = mpz_sizeinbase(x, 2);

    assert(n1 > m);
    if (n1 > m + 1 + (n+3)/4)
      {
	trace("hgcd(%d): n1 unexpectedly large, got %d, expected <= %d",
	      count, n1, m + 1 + (n+3)/4);
      }
    
    /* Determine how many bits to split off in the recursive call. */
    l = 2*m - n1 + 1;

    assert(l < n1);
    assert(l + (n1 + 1 - l)/2 > m);
	
    assert(mpz_cmp(x, y) > 0);	

    if (mpz_sizeinbase(y, 2) > l + (n1 + 1 - l) / 2)
      {
#if 0
	trace("hgcd(%d) reduced: x = %Zd, y = %Zd\n", count, x, y);
#endif
	mpz_fdiv_q_2exp(x1, x, l);
	mpz_fdiv_q_2exp(y1, y, l);

	/* schoenhage_array_init(&S); */
      
	hgcd(&S, x1, y1);
	check_hgcd(&S, x1, y1);
	schoenhage_array_vector(x, y, &S, x, y);

	fix(&S, x, y);
      }
    else
      schoenhage_array_identity(&S);

    schoenhage_array_mult(R, &S, R);
    schoenhage_array_clear(&S);
    
    assert(mpz_sizeinbase(x, 2) > m);
    if (mpz_sizeinbase(y, 2) <= m)
      goto done;

    assert(mpz_sizeinbase(y, 2) <= m + 2); 
    while (mpz_sizeinbase(y, 2) > m)
      {
	mpz_fdiv_qr(q, r, x, y);
	mpz_swap(r, x);
	mpz_swap(x, y);
	schoenhage_array_step_left(R, q);
      }
    
  done:
    assert(mpz_sizeinbase(x, 2) > m);
    assert(mpz_sizeinbase(y, 2) <= m);
    
    mpz_clear(x);
    mpz_clear(y);    
    mpz_clear(x1);
    mpz_clear(y1);
    mpz_clear(q);
    mpz_clear(r);
  }
  
#if 0
  trace("hgcd(%d): mult, R = [ %Zd, %Zd\n"
	"                      %Zd, %Zd ]\n", count,
	R->m[0][0], R->m[0][1], R->m[1][0], R->m[1][1]);
#endif
}

void
hgcd_slow(struct schoenhage_array *R, const mpz_t x, const mpz_t y)
{
  /* We want the product corresponding to remainders of size larger
   * than this. */
  unsigned m = (mpz_sizeinbase(x, 2) + 1) / 2;
  mpz_t a;
  mpz_t b;
  mpz_t r;
  mpz_t q;

  assert(mpz_cmp(x, y) > 0);
  
  trace("hgcd_slow:");

  schoenhage_array_identity(R);

  mpz_init(r);
  mpz_init(q);
  mpz_init_set(a, x);
  mpz_init_set(b, y);

  /* Through out this loop, (a;b) = R (original_a;original_b) */

  while (mpz_sizeinbase(b, 2) > m)
    {
      mpz_fdiv_qr(q, r, a, b);

      trace("(q,r) = (%Zd, %Zd) ", q, r);
      schoenhage_array_step_left(R, q);  
      
      mpz_swap(a, r);
      mpz_swap(a, b);
    }
  trace("\n");
  
  mpz_clear(r);
  mpz_clear(q);
  mpz_clear(a);
  mpz_clear(b);
}

/* Must be called with a >= b > 0. Modifies a and b */
void
fgcd(mpz_t r, mpz_t a, mpz_t b)
{
  for (;;)
    {
      struct schoenhage_array R;
      unsigned before;
      
      if (mpz_size(b) < SCHOENHAGE_LIMIT)
	{
	  mpz_gcd(r, a, b);
	  return;
	}

      trace("fgcd: a = %Zd, b = %Zd\n", a, b);

      before = mpz_sizeinbase(a, 2);
      schoenhage_array_init(&R);

      if (mpz_sizeinbase(b, 2) > (mpz_sizeinbase(a,2) + 1)/2)
	{
	  hgcd(&R, a, b);
	  check_hgcd(&R, a, b);
	  
	  /* Updates a and b */
	  schoenhage_array_vector(a, b, &R, a, b);

	  schoenhage_array_clear(&R);
	}
      if (!normalize(r, a, b))
	return;

      /* One euclidean step */
      mpz_mod(a, a, b);
      if (!mpz_sgn(a))
	{
	  mpz_set(r, b);
	  return;
	}
      mpz_swap(a, b);

      trace("fgcd: size reduction: before: %d, after: %d\n",
	    before, mpz_sizeinbase(a, 2));
    }
}

void
schoenhage(mpz_t r, const mpz_t x, const mpz_t y)
{
  int fixed = 0;
  mpz_t a;
  mpz_t b;
  mpz_t t;
  
  mpz_init_set(a, x);
  mpz_init_set(b, y);
  mpz_init(t);
  
  /* Ensure a >= b > 0 */
  if (!normalize(r, a, b))
    return;
  
  fgcd(r, a, b);

  /* fgcd may have introduced spurious factors */
  mpz_mod(t, x, r);
  if (mpz_sgn(t))
    {
      fixed = 1;
      trace("schoenhage: r= %Zd does not divide x. Reducing...\n", r);
      mpz_gcd(r, r, t);
    }

  mpz_mod(t, x, r);
  if (mpz_sgn(t))
    {
      fixed = 1;
      trace("schoenhage: r = %Zd does not divide y. Reducing...\n", r);
      mpz_gcd(r, r, t);
    }
  if (fixed)
    trace("after: r = %Zd\n", r);
}
