#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "schoenhage.h"

#define ASSERT(x) \
do { if (!(x)) abort(); } while (0)

#if 0
static void
try_egcd(unsigned size, unsigned x, unsigned y)
{
  struct schoenhage_array R;
  struct schoenhage_array S;
  unsigned a, b;
  mpz_t s;
  mpz_t t;
  
  mpz_init(s);
  mpz_init(t);
  
  schoenhage_array_init(&R);
  schoenhage_array_identity(&R);

  a = x; b = y;

  for (;;)
    {
      unsigned q = a / b;
      unsigned r = a % b;

      trace("try_egcd: a = %d, b = %d, q = %d, r = %d\n",
	    a, b, q, r);
      
      schoenhage_array_step_ui(&R, q);
      
      trace(" R = [ %Zd, %Zd ; %Zd, %Zd ]\n",
	    R.m[0][0], R.m[0][1], R.m[1][0], R.m[1][1]);

      /* Stop here, to match what egcd does */
      if (r < 2)
	break;
      
      mpz_mul_si(s, R.m[0][0], x);
      mpz_mul_si(t, R.m[0][1], y);
      mpz_add(s, s, t);

      ASSERT(!mpz_cmp_si(s, b));

      a = b; b = r;
    }

  schoenhage_array_init(&S);
  egcd(&S, size, x, y);

  trace_array(" S", &S);
  
  ASSERT(size || schoenhage_array_equal(&R, &S));
  
  schoenhage_array_clear(&R);
  schoenhage_array_clear(&S);
  mpz_clear(s);
  mpz_clear(t);
}
#endif

static void
try_hgcd(const char *sa, const char *sb)
{
  struct schoenhage_array R;
  struct schoenhage_array S;

  mpz_t a;
  mpz_t b;

  schoenhage_array_init(&R);
  schoenhage_array_init(&S);
  mpz_init(a);
  mpz_init(b);

  mpz_set_str(a, sa, 10);
  mpz_set_str(b, sb, 10);

  trace("try_hgcd: a = %Zd, b = %Zd\n", a, b);
  hgcd_slow(&R, a, b);
  check_hgcd(&R, a, b);
  hgcd(&S, a, b);
  check_hgcd(&S, a, b);

  trace_array(" R", &R);
  trace_array(" S", &S);

  if (!schoenhage_array_equal(&R, &S))
    trace("try_hgcd: Mismatch\n");
  /* ASSERT(schoenhage_array_equal(&R, &S)); */
  
  schoenhage_array_clear(&R);
  schoenhage_array_clear(&S);
  mpz_clear(a);
  mpz_clear(b);
}

static void
try_hgcd_fib(unsigned n)
{
  struct schoenhage_array R;
  struct schoenhage_array S;

  mpz_t a;
  mpz_t b;

  schoenhage_array_init(&R);
  schoenhage_array_init(&S);
  mpz_init(a);
  mpz_init(b);

  mpz_fib2_ui(a, b, n);

  trace("try_hgcd: n = %d, a = %Zd, b = %Zd\n", n, a, b);
  hgcd_slow(&R, a, b);
  trace_array(" R", &R);
  check_hgcd(&R, a, b);
  hgcd(&S, a, b);
  trace_array(" S", &S);
  check_hgcd(&S, a, b);

  ASSERT(schoenhage_array_equal(&R, &S));
  
  schoenhage_array_clear(&R);
  schoenhage_array_clear(&S);
  mpz_clear(a);
  mpz_clear(b);
}

static void
try_fgcd(unsigned n)
{
  mpz_t a;
  mpz_t b;
  mpz_t r;
  
  mpz_init(a);
  mpz_init(b);
  mpz_init(r);

  trace("\ntry_fgcd: n = %d\n", n);
  
  mpz_fib2_ui(a, b, n);
  schoenhage(r, a, b);

  trace("schoenhage(%Zd, %Zd) = %Zd\n",
	a, b, r);
  
  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(r);
}

#define RANDOM(r, state, size) do {		\
  mpz_urandomb((r), (state), (size));		\
  mpz_setbit(r, (size)-1);			\
} while (0)

static void
try_random(void)
{
  unsigned size;
  
  gmp_randstate_t state;
  time_t now;

  /* now = time(NULL); */
  now = 1055513922;

  trace("try_random: Seeding with %ld\n", (long) now);
  
  gmp_randinit_default(state);
  gmp_randseed_ui(state, now);

  for (size = 10; size < 1000; size += 3)
    {
      mpz_t a;
      mpz_t b;
      mpz_t c;
      mpz_t g;
      mpz_t t;

      mpz_init(a);
      mpz_init(b);
      mpz_init(c);
      mpz_init(g);
      mpz_init(t);

      RANDOM(a, state, size + 17);
      RANDOM(b, state, size + 47);
      RANDOM(t, state, size);

      mpz_mul(a, a, t);
      mpz_mul(b, b, t);
      mpz_gcd(g, a, b);

      trace("try_random: gcd(%Zd, %Zd) = %Zd\n",
	    a, b, g);

      schoenhage(t, a, b);
      if (mpz_cmp(t, g))
	{
	  trace("try_random: Incorrect result: %Zd\n");
	  abort();
	}
      
      mpz_clear(a);
      mpz_clear(b);
      mpz_clear(c);
      mpz_clear(g);
      mpz_clear(t);
    }
  
      
  gmp_randclear(state);
}

int
main(int argc, char **argv)
{
  unsigned n;
  unsigned i;

  mpz_t a;
  mpz_t b;
  mpz_t r;
  
  mpz_init(a);
  mpz_init(b);
  mpz_init(r);

  n = 0;
  if (argc > 1)
    n = atoi(argv[1]);
  
  /* normalize */
  mpz_set_si(a, -7);
  mpz_set_si(b, 3);
  
  ASSERT(normalize(r, a, b));
  ASSERT(!mpz_cmp_si(a, 7));
  ASSERT(!mpz_cmp_si(b, 3));

  mpz_set_si(a, 0);
  ASSERT(!normalize(r, a, b));
  ASSERT(!mpz_cmp_si(r, 3));

#if 0
  /* schoenhage_array_step_ui */
  try_egcd(0, 268, 87);
  try_egcd(0, 270, 87);
  /* Problematic case for gcd(fib(100), fib(99)) */
  try_egcd(8, 87833, 6075);
#endif
  
  for (i = 6; i<17; i++)
    try_hgcd_fib(i);

  try_hgcd("133304756", "12777598");
  try_hgcd("4368130264137", "418696360162");
  
  try_hgcd("30019063369843142", "9258668778823265"); /* !!! */
  
  try_hgcd("133304756", "12777598");
  try_hgcd("223659451", "68982457");
  try_hgcd("8259214276", "5104443020");
  try_hgcd("70946110418820658019", "43846831676944464992");
  try_hgcd("14901987450837196642636732651392",
	   "11108588010711814328108699732800");
  try_hgcd("1225680876558579559649902923502714694035764",
	   "812484106832574967003225854678869801985244");
  
  if (n)
    try_fgcd(n);
  else
    {
      for (i = 1; i<10; i++)
	try_fgcd(i);
  
      for( ; i < 100000; i *= 10)
	try_fgcd(i);
    }

  try_random();
  
  mpz_clear(a);
  mpz_clear(b);
  mpz_clear(r);

  return EXIT_SUCCESS;
}
